use crate::prelude::*;
use anyhow::Result;
use utilities::{
    impl_reprc,
    prelude::cgmath::{Matrix4, SquareMatrix},
};

use super::camera::Camera;

use std::sync::Arc;

impl_reprc!(
    #[derive(Debug, Copy)]
    pub struct CameraProperties {
        #[assume_reprc]
        view: Matrix4<f32>,
        #[assume_reprc]
        proj: Matrix4<f32>,
        #[assume_reprc]
        vp: Matrix4<f32>,

        #[assume_reprc]
        inv_view: Matrix4<f32>,
        #[assume_reprc]
        inv_proj: Matrix4<f32>,
    }
);

#[derive(Clone)]
pub struct View {
    proj: cgmath::Matrix4<f32>,
    camera: Camera,

    desc_set: Option<Arc<DescriptorSet>>,
    buffer: Arc<Buffer<CameraProperties>>,
}

impl View {
    /// creates a view handle
    pub fn new(
        device: &Arc<Device>,
        proj: cgmath::Matrix4<f32>,
        fov: f32,
        create_descriptor: bool,
    ) -> Result<View> {
        let buffer = Buffer::builder()
            .set_usage(VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT)
            .set_memory_usage(MemoryUsage::CpuOnly)
            .set_size(1)
            .build(device.clone())?;

        let desc_set = if create_descriptor {
            let desc_layout = Self::desc_layout(device)?;

            let desc_pool = DescriptorPool::builder()
                .set_layout(desc_layout)
                .build(device.clone())?;

            let desc_set = DescriptorPool::prepare_set(&desc_pool).allocate()?;

            desc_set.update(&[DescriptorWrite::uniform_buffers(0, &[&buffer])])?;

            Some(desc_set)
        } else {
            None
        };

        Ok(View {
            proj,
            camera: Camera::new(fov),

            desc_set,
            buffer,
        })
    }

    pub fn frustum(&self) -> ViewFrustum {
        let view_matrix = self.camera.old_view_matrix();
        let proj_matrix = self.proj();

        let mvp = proj_matrix * view_matrix;
        let inv_mvp = mvp.invert().unwrap();

        ViewFrustum::new(inv_mvp)
    }

    pub fn new_proj(view: &View, proj: cgmath::Matrix4<f32>, fov: f32) -> Result<View> {
        let mut new_view = View {
            proj,
            camera: view.camera.clone(),

            desc_set: view.desc_set.clone(),
            buffer: view.buffer.clone(),
        };

        new_view.camera.set_fov(fov);
        new_view.update_buffer()?;

        Ok(new_view)
    }

    /// returns the layout used to present this view
    pub fn desc_layout(device: &Arc<Device>) -> Result<Arc<DescriptorSetLayout>> {
        Ok(DescriptorSetLayout::builder()
            .add_layout_binding(0, VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, VK_SHADER_STAGE_ALL, 0)
            .build(device.clone())?)
    }

    /// returns the camera handle to read this view
    pub fn camera(&self) -> &Camera {
        &self.camera
    }

    /// returns the camera handle to configure this view
    pub fn camera_mut(&mut self) -> &mut Camera {
        &mut self.camera
    }

    /// updates the projection matrix
    pub fn set_proj(&mut self, proj: Matrix4<f32>) {
        self.proj = proj;
    }

    /// returns the projection matrix
    pub fn proj(&self) -> &Matrix4<f32> {
        &self.proj
    }

    /// updates the vulkan buffer with current configuration
    /// needs to be called after updating the camera paramter
    pub fn update_buffer(&mut self) -> Result<()> {
        let camera = &mut self.camera;
        let proj = self.proj;

        let cam = CameraProperties {
            view: camera.view_matrix(),
            proj: proj.clone(),
            vp: proj.clone() * camera.view_matrix(),

            inv_proj: proj.invert().unwrap(),
            inv_view: camera.view_matrix().invert().unwrap(),
        };

        Ok(self.buffer.fill(&[cam])?)
    }

    /// returns the descriptor set where this buffer is bound to
    pub fn descriptor_set(&self) -> &Arc<DescriptorSet> {
        self.desc_set.as_ref().unwrap()
    }

    /// returns the buffer handle itself
    pub fn buffer(&self) -> &Arc<Buffer<CameraProperties>> {
        &self.buffer
    }
}
